#/*******************************************************************************
# * (C) Copyright 2003/2004 by Stanislav Soldatov  <stass@graphics.cs.msu.su>   *
# *******************************************************************************
# * $Id: cvsfs.cpp,v 1.9 2004/01/27 22:54:12 stass Exp $
# *******************************************************************************
# * 
# *******************************************************************************
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# *******************************************************************************/

#include "stdafx.h"
#include "PlugCVS.h"
#include "cvsfs.h"

#include <map>

ViewOperation_enum ShowViewMenu(const char* filename,int sel_files);

static KeyBarTitles g_keybar;
static char *tit_m0[5]={"Name","Opt","WRev","Status","Date"};
static char *tit_m1[6]={"Name","S","Name","S","Name","S"};
static char *tit_m3[5]={"Name","Status","WorkRev","Opt","Size"};
static char *tit_m4[3]={"S","Name","Revision"};
static char *tit_m5[7]={"Name","Status","Size","Modified","WorkRev","RepRev","Opt"};
static PanelMode g_panelmodes[6]=
{
  {"NOM,C3,C1,Z,D","0,3,4,0,5",tit_m0,0,0,1,0,"NOR,Z","0,0"},
  {"NO,C0,NO,C0,NO,C0","0,1,0,1,0,1",tit_m1,0,0,0,0,"NOR,C1,S,D,T","0,4,5,5,5"},
  {"NO,C0,NO,C0","0,1,0,1",tit_m1,0,0,0,0,"NOR,C1,S,D,T","0,4,5,5,5"},
  {"NO,Z,C1,C3,S","0,10,5,3,5",tit_m3,0,0,1,0,"C0,NOR,D,T","1,0,0,5"},
  {"C0,NO,C1","1,0,8",tit_m4,0,0,0,0,"NOR,C1,S,D,T","0,4,5,5,5"},
  {"NO,Z,S,DM,C1,C2,C3","0,0,6,14,7,7,3",tit_m5,1,0,1,0,"C0,NR","1,0"}
};


typedef map<string,PluginPanelItem*> item_map;



//===================================================================
//= Function name	: GetHistory
//= Description   : 
//= Return type   : void 
//===================================================================

void GetHistory(item_map& files,const char* dir=0)
{
    chdir(dir);

    const char* args[5];
    args[0]="cvs";
    args[1]="-nq";
    args[2]="update";
    args[3]="-l";
    args[4]=0;
  
    const char* fname = _tempnam(getenv("TEMP"), "cvsplug" );
    if(!fname)
      return;
  
    int iStatus = RunCVS(args, fname, false);
    //    
    //TODO:   box
    if(iStatus)
      ShowLogs(0 ,fname,true);
    else
    {
      // 
      FILE *out = fopen(fname, "rt");
      if (!out)
        goto runaway;

      char buf[312];
      char filename[256];
      char type='*';

      while (!feof(out) && !ferror(out))
        if (fgets(buf, sizeof(buf), out))
          if (sscanf(buf, " %c%*1[ ]%s",&type, filename)>=2)
            {
              item_map::const_iterator f=files.find(filename);
              if(f!=files.end())
              {
                f->second->Description=new char[2];
                f->second->Description[0]=type;
                f->second->Description[1]=0;
              }
            }

      fclose(out);

    }
  runaway:
    ///   
    _unlink(fname);

}

void FillCVSFileInfo(PluginPanelItem* fi,
                     const char* status,const char* stag,const char* sdate,const char* sopt,const char* mergefrom,
                     char type,const char* wrev,const char* rrev,const char* expopt)
{
  char* desc=fi->Description=new char[256];
  strcpy(desc,status);
  if(stag[0])
  { strcat(desc," STag: ");strcat(desc,stag); }
  if(sdate[0])
  { strcat(desc," SDate: ");strcat(desc,sdate); }
  if(sopt[0])
  { strcat(desc," SOpt: ");strcat(desc,sopt); }
  if(mergefrom[0])
  { strcat(desc," Merge: ");strcat(desc,mergefrom); }
  
  char ** data=new char*[4];
  memset(data,0,sizeof(char*)*4);

  data[0]=new char[2];
  data[0][0]=type;
  data[0][1]=0;
  if(wrev[0])
    data[1]=strdup(wrev);
  if(rrev[0])
    data[2]=strdup(rrev);
  if(expopt[0])
    data[3]=strdup(expopt);

  fi->CustomColumnData=data;
  fi->CustomColumnNumber=4;

  fi->UserData=type;

  fi->FindData.dwFileAttributes&=~(FILE_ATTRIBUTE_ENCRYPTED|FILE_ATTRIBUTE_COMPRESSED);
  if(type=='U')
    fi->FindData.dwFileAttributes|=FILE_ATTRIBUTE_ENCRYPTED;
  if(type=='M'||type=='A'||type=='R')
    fi->FindData.dwFileAttributes|=FILE_ATTRIBUTE_COMPRESSED;

}


//===================================================================
//= Function name	: GetStatus
//= Description   : 
//= Return type   : void 
//===================================================================

void GetStatus(item_map& files,const char* dir=0)
{
    chdir(dir);

    const char* args[5];
    args[0]="cvs";
    args[1]="-q";
    args[2]="status";
    args[3]="-l";
    args[4]=0;
  
    const char* fname = _tempnam(getenv("TEMP"), "cvsplug" );
    if(!fname)
      return;
  
    int iStatus = RunCVS(args, fname, false);
    //    
    //TODO:   box
    if(iStatus)
    { ShowLogs(0 ,fname,true); }
    else
    {
      // 
      FILE *out = fopen(fname, "rt");
      if (!out)
        goto runaway;

      char buf[312];
      char filename[128];
      char status[48];
      char wrev[24],rrev[24];
      char expopt[16],stag[32],sdate[32],sopt[32],mergefrom[32];
      char type=0;

      item_map::const_iterator fi;

      filename[0]=0;status[0]=0;wrev[0]=0;rrev[0]=0;type=0;
      expopt[0]=0;stag[0]=0;sdate[0]=0;sopt[0]=0;mergefrom[0]=0;

      while (!feof(out) && !ferror(out))
      {
        if(fgets(buf, sizeof(buf), out)==NULL) break;

        if (strncmp(buf, "===============",15)==0)
        {
          if(filename[0])
          {
            fi=files.find(filename);
            if(fi!=files.end())
              FillCVSFileInfo(fi->second,
                              status,stag,sdate,sopt,mergefrom,
                              type,wrev,rrev,expopt);
          };
          filename[0]=0;status[0]=0;wrev[0]=0;rrev[0]=0;type=0;
          expopt[0]=0;stag[0]=0;sdate[0]=0;sopt[0]=0;mergefrom[0]=0;
          continue;
        };

        if(sscanf(buf, " File: %s Status: %[^\n]",filename,status)==2)
        {
          if(strcmp(status,"Up-to-date")==0)
            type=' ';
          else if(strcmp(status,"Locally Modified")==0)
		  {
			strcpy(status,"Modified");
            type='M';
		  }
          else if(strcmp(status,"Needs Patch")==0)
            type='U';
          else if(strcmp(status,"Locally Added")==0)
		  {
			strcpy(status,"Added");
            type='A';
		  }
          else
            type=status[0];
          continue;
        }

        if(sscanf(buf, " Working revision: %s",wrev)==1) 
          continue;

        if(sscanf(buf, " Repository revision: %s",rrev)==1)
          continue;

        if(sscanf(buf, " Expansion option: %s",expopt)==1)
        {
          if(strcmp(expopt,"(default)")==0)
            expopt[0]=0;
          continue;
        };

        if(sscanf(buf, " Sticky Tag: %s",stag)==1)
        {
          if(strcmp(stag,"(none)")==0)
            stag[0]=0;
          continue;
        };
        
        if(sscanf(buf, " Sticky Date: %s",sdate)==1)
        {
          if(strcmp(sdate,"(none)")==0)
            sdate[0]=0;
          continue;
        };
        
        if(sscanf(buf, " Sticky Options: %s",sopt)==1)
        {
          if(strcmp(sopt,"(none)")==0)
            sopt[0]=0;
          continue;
        };
        
        if(sscanf(buf, " Merge From: %s",mergefrom)==1)
        {
          if(strcmp(mergefrom,"(none)")==0)
            mergefrom[0]=0;
          continue;
        };
      };

      if(filename[0])
      {
        fi=files.find(filename);
        if(fi!=files.end())
          FillCVSFileInfo(fi->second,
                          status,stag,sdate,sopt,mergefrom,
                          type,wrev,rrev,expopt);
      };

      fclose(out);

    }
  runaway:
    ///   
    _unlink(fname);
}



//===================================================================
//= Function name	: _export GetFindData
//= Description   : 
//= Return type   : int WINAPI 
//===================================================================

int WINAPI _export GetFindData( HANDLE hPlugin, PluginPanelItem **pitems, int *item_number, int opmode)
{
   
  CVSFS* pFS=(CVSFS*)hPlugin;


  //      win
  string mask=pFS->m_dir;
  mask+="\\*.*";


  *item_number=0;
  WIN32_FIND_DATA find_data;

  HANDLE find_handle=FindFirstFile(mask.c_str(),&find_data);
  if(find_handle==INVALID_HANDLE_VALUE)
    return TRUE;


  item_map files;
  string current_dir_s=pFS->m_dir;
  current_dir_s+="\\";
  const char* current_dir=current_dir_s.c_str();
  int current_dir_len=current_dir_s.length();


  do
  {
    //  
    if(strcmp(find_data.cFileName,".")==0
      ||strcmp(find_data.cFileName,"..")==0
      ||strcmp(find_data.cFileName,"CVS")==0)
      continue;
    PluginPanelItem* it=new PluginPanelItem;
    memset(it,0,sizeof(PluginPanelItem));
    it->FindData.dwFileAttributes = find_data.dwFileAttributes;
    it->FindData.ftCreationTime   = find_data.ftCreationTime;
    it->FindData.ftLastAccessTime = find_data.ftLastAccessTime;
    it->FindData.ftLastWriteTime  = find_data.ftLastWriteTime;
    it->FindData.nFileSizeHigh    = find_data.nFileSizeHigh;
    it->FindData.nFileSizeLow     = find_data.nFileSizeLow;
    it->FindData.dwReserved0      = find_data.dwReserved0;
    it->FindData.dwReserved1      = find_data.dwReserved1;

    strcpy(it->FindData.cFileName,current_dir);
    strcpy(it->FindData.cFileName+current_dir_len,find_data.cFileName);

    strcpy(it->FindData.cAlternateFileName,find_data.cAlternateFileName);

    string name=find_data.cFileName;
    files.insert(item_map::value_type(name,it));

  }while(FindNextFile(find_handle,&find_data));


//  cvs
  if(pFS->is_cvsdir)
  {
    GetStatus(files,pFS->m_dir.c_str());
  };

//  
  for(item_map::const_iterator k=files.begin();k!=files.end();k++)
    if(k->second->CustomColumnData==NULL 
      && !(k->second->FindData.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY) )
    {
      k->second->Description=strdup("Unknown");
      char ** data=new char*[1];
      data[0]=strdup("?");
      k->second->CustomColumnData=data;
      k->second->CustomColumnNumber=1;
      k->second->UserData='?';
    };
  //   
  int size=files.size();
  PluginPanelItem* items=new PluginPanelItem[size];

  item_map::iterator i=files.begin();
  for(int j=0;i!=files.end();i++,j++)
  {
    items[j]=*(i->second);
    delete i->second;
    i->second=0;
  };

  *pitems=items;
  *item_number=size;


  return TRUE;
};


//===================================================================
//= Function name	:  _export  FreeFindData
//= Description   : 
//= Return type   : void WINAPI 
//===================================================================

void WINAPI  _export  FreeFindData( HANDLE hPlugin, struct PluginPanelItem *items, int n)
{
  CVSFS* pFS=(CVSFS*)hPlugin;

  if(n && items)
  {
    for(int i=0;i<n;i++)
    {
      if(items[i].Description)
        delete[] items[i].Description;
      if(items[i].CustomColumnData)
      {
        for(int j=0;j<items[i].CustomColumnNumber;j++)
          delete[] items[i].CustomColumnData[j];
        delete[] items[i].CustomColumnData;
      }
    }

    delete[] items;
  };
};

static InfoPanelLine g_infolines[3]=
{
  {"CVS Roots","",1},
  {"Default Root","",0},
  {"Current Root","",0}
};


//===================================================================
//= Function name	: _export GetOpenPluginInfo
//= Description   : 
//= Return type   : void WINAPI 
//===================================================================

void WINAPI _export GetOpenPluginInfo( HANDLE hPlugin, struct OpenPluginInfo *out_pInfo)
{
  CVSFS* pFS = (CVSFS*)hPlugin;
  string sTitle;
  char  *pcString;

  out_pInfo->StructSize = sizeof(OpenPluginInfo);
  out_pInfo->Flags = OPIF_ADDDOTS | OPIF_USEHIGHLIGHTING | OPIF_USESORTGROUPS | OPIF_REALNAMES;

  sTitle ="CVS:";
  sTitle += pFS->m_dir;
  strcpy(pFS->m_title, sTitle.c_str());

  out_pInfo->PanelTitle = pFS->m_title;
  out_pInfo->CurDir = (char*)pFS->m_dir.c_str();
  out_pInfo->ShortcutData = pFS->m_title;
  out_pInfo->KeyBar = &g_keybar;
  out_pInfo->PanelModesArray = g_panelmodes;
  out_pInfo->PanelModesNumber = sizeof(g_panelmodes) / sizeof(g_panelmodes[0]);
  out_pInfo->StartPanelMode = '0' + pFS->m_start_mode;


  pcString = getenv("CVSROOT");
  if (pcString)
    strcpy(g_infolines[1].Data, pcString);
  else
    strcpy(g_infolines[1].Data, "not defined");

  strcpy(g_infolines[2].Data, pFS->m_cvs_root.c_str());

  out_pInfo->InfoLines = g_infolines;
  out_pInfo->InfoLinesNumber = sizeof(g_infolines) / sizeof(g_infolines[0]);
};


//===================================================================
//= Function name	: _export ClosePlugin
//= Description   : 
//= Return type   : void WINAPI 
//===================================================================

void WINAPI _export ClosePlugin(HANDLE hPlugin)
{
    CVSFS* pFS=(CVSFS*)hPlugin;

    delete pFS;
}


//===================================================================
//= Function name	: _export SetDirectory
//= Description   : 
//= Return type   : int WINAPI 
//===================================================================

int WINAPI _export SetDirectory( HANDLE hPlugin, const char *dir, int OpMode)
{
  CVSFS* pFS=(CVSFS*)hPlugin;

  if(!dir || strcmp(dir,"\\")==0)
  {
    int n = pFS->m_dir.find_first_of('\\');
    if(n!=string::npos)
    {
      pFS->m_dir.resize(n);
      pFS->ChangeDir(pFS->m_dir);
    };
  }
  else if(strchr(dir,'\\')!=NULL) //full dir name
  {
      pFS->ChangeDir(dir);
  }
  else if(strcmp(dir,"..")==0)
  {
    int n=pFS->m_dir.find_last_of('\\');
    if(n!=string::npos)
    {
      pFS->m_dir.resize(n);
      pFS->ChangeDir(pFS->m_dir);
    };
  }
  else
  {
    pFS->m_dir.append("\\");
    pFS->m_dir.append(dir);
    pFS->ChangeDir(pFS->m_dir);
  };

  //if(!pFS->is_cvsdir)
  //  g_Info.Control(pFS,FCTL_CLOSEPLUGIN,(void*)pFS->m_dir.c_str());
  return TRUE;
};


//===================================================================
//= Function name	: _export ProcessKey
//= Description   : Itercept far key processing
//= Return value  : TRUE to continue Far key processing
//===================================================================

int WINAPI _export ProcessKey( HANDLE hPlugin, int Key, unsigned int ControlState)
{
   CVSFS* fs=(CVSFS*)hPlugin;

   switch(Key)
   {
   case VK_ESCAPE:
     if(ControlState==PKF_SHIFT)
     {  
       g_Info.Control(fs,FCTL_CLOSEPLUGIN,(void*)fs->m_dir.c_str());  
       return TRUE;
     }
     break;

   case VK_RETURN:
     if(ControlState==PKF_CONTROL)
     {  
       return PasteName();
     }
     break;

   case VK_F3:
     if(ControlState==0 && fs->is_cvsdir)
        return View(); 
     break;

   }
   // do not process
   return FALSE;
}


//===================================================================
//= Function name	: Browse
//= Description   : 
//= Return type   : HANDLE 
//===================================================================
HANDLE Browse(const char *path)
{
  string dir;

  PanelInfo        PInfo;

  ///    
  g_Info.Control(INVALID_HANDLE_VALUE, FCTL_GETPANELINFO, &PInfo);

  ///   ,   - 
  if (PInfo.PanelType != PTYPE_FILEPANEL)
    return INVALID_HANDLE_VALUE;

  if(path)
    dir=path;
  else
    dir=PInfo.CurDir;
    
 
  return new CVSFS(dir,PInfo.ViewMode);
}


//===================================================================
//= Function name	: CVSFS::ChangeDir
//= Description   : 
//= Return type   : void 
//===================================================================

void CVSFS::ChangeDir(const string& _dir)
{
  is_cvsdir=false;
  m_cvs_root="";

  string tmp=m_dir=_dir;
  tmp+="\\CVS";

   struct _stat finfo;
  if(_stat(tmp.c_str(),&finfo)==-1 || !(finfo.st_mode&_S_IFDIR))
    return;

  is_cvsdir=true;
  
  tmp+="\\Root";
  FILE *root= fopen(tmp.c_str(), "rt");
  if (!root)
    return;

  char buf[348];

  if(fgets(buf, sizeof(buf), root))
  {
    *strchr(buf,'\n')=0;
    m_cvs_root=buf;
  }

  fclose(root);
  
};


//===================================================================
//= Function name	: View
//= Description   : 
//= Return type   : bool 
//===================================================================

bool View()
{
  PanelInfo        panel_info;
  std::vector <std::string> sel_files;

  ///    
  g_Info.Control(INVALID_HANDLE_VALUE, FCTL_GETPANELINFO, &panel_info);

  ///   ,   - 
  if (panel_info.PanelType != PTYPE_FILEPANEL)
    return false;

  //   -  
  if (!panel_info.ItemsNumber)
    return false;

  const char* cursor_file = 0;

  PluginPanelItem* cursor_item=&panel_info.PanelItems[panel_info.CurrentItem];
  char cursor_item_type=(char)cursor_item->UserData;
  cursor_file=cursor_item->FindData.cFileName;

  ///     -  
  for (int i = 0; i < panel_info.SelectedItemsNumber; i++)      
    if(panel_info.SelectedItems[0].Flags & PPIF_SELECTED)
	{
	  const char* name=strrchr(panel_info.SelectedItems[i].FindData.cFileName,'\\');
	  if(name) 
		  name++;
	  else
		  name=panel_info.SelectedItems[i].FindData.cFileName;
	  sel_files.push_back(name);
	}

  const char* short_name=0;
  if(cursor_file) short_name=strrchr(cursor_file,'\\');
  if(short_name) short_name++;

  int sel_num=sel_files.size();

  if(sel_num==0 && cursor_item_type=='?')
    return false;

  ViewOperation_enum op=ShowViewMenu(short_name,sel_num);
  
  vector <const char *> args;
  ///   
  const arg_num=2;
  args.resize(arg_num);
  args[0] = "cvs";
  
  switch(op)
  {
  case VOP_LOG:
	if(!short_name)
		return true;
    args[1]="log";
    args.push_back(short_name);
    break;

  case VOP_DIFF:
	if(!short_name)
		return true;
    args[1]="diff";
    args.push_back(short_name);
    break;

  case VOP_LOGMANY:
    args[1]="log";
    break;

  case VOP_DIFFMANY:
    args[1]="diff";
    break;

  case VOP_VIEW:
    return false;
  case VOP_INVALID:
  default:
    return true;
  };

  if(op==VOP_LOGMANY || op== VOP_DIFFMANY)
  {
    ///    
    args.resize(arg_num + sel_files.size());

    for (int i = 0; i < sel_files.size(); i++)  
      args[i + arg_num] = sel_files[i].c_str();
  };

  args.push_back(NULL);


  const char* fname = _tempnam(getenv("TEMP"), "cvsplug" );
  if(!fname)
    return false;

  ///  CVS
  int iStatus = RunCVS(&args[0], fname, true);
  if(op==VOP_DIFF || op==VOP_DIFFMANY)
    iStatus=0;

  ///  
  ShowLogs(iStatus, fname, true);

  ///   
  _unlink(fname);

  return true;
}


//===================================================================
//= Function name	: PasteName
//= Description   : Overloads Ctrl-Enter by pasting name withoout path
//= Return type   : bool 
//===================================================================

bool PasteName()
{
  PanelInfo        panel_info;

  ///    
  g_Info.Control(INVALID_HANDLE_VALUE, FCTL_GETPANELINFO, &panel_info);

  //   -  
  if (!panel_info.ItemsNumber)
    return false;


  PluginPanelItem* cursor_item=&panel_info.PanelItems[panel_info.CurrentItem];

  const char* cursor_file = cursor_item->FindData.cFileName;

  if(!cursor_file)
    return false;

  char buf[MAX_PATH];

  strcpy(buf,g_FSF.PointToName(cursor_file));
  strcat(buf," ");

  g_Info.Control(INVALID_HANDLE_VALUE,FCTL_INSERTCMDLINE,buf);  

  return true;
};